# Copyright 2017-2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""IO functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os


def GetResource(name, mode='rb'):
  with open(name, mode=mode) as f:
    return f.read()


def GetResourceFilename(name, mode='rb'):
  del mode  # Unused.
  return name


def WalkResources(path):
  return os.walk(path)


GetResourceAsFile = open  # pylint: disable=invalid-name
